#!/usr/bin/perl

# WebVNC:
#
# Small Perl-Daemon that checks which CTs and VMs have VNC enabled.
# For each CT or VM with enabled VNC it will set up a websocket that
# connects the VNC port to a publically exposed Web-Accessibble port
# for usage with noVNC.

# Debugging switch - logs to /var/log/messages:
$DEBUG = "1";
if ($DEBUG) {
    use Sys::Syslog qw( :DEFAULT setlogsock);
}

use Time::HiRes qw(usleep);
use JSON;
use Data::Dumper;
use Systemd::Daemon qw( -hard notify );

my $sleep = ($ENV{WATCHDOG_USEC} // 2_000_000) / 2;
$| = 1;
notify( READY => 1 );

my %vnc_users;

$runner = '0';
$firstrun = '1';

# Initial startup run:
&debug_msg("Startup. \n");
&vzlist;
&parse_vzlist;
&launcher;

# Loop for further executions and for Systemd-Watchdog:
while (1) {

    &debug_msg("Sending Systemd Watchdog event. \n");

    # Bark, watchdog, bark!
    notify( WATCHDOG => 1 );

    # Execute Websocket starter:
    if ($firstrun eq "1") {
        # Skip exec on firstrun.
    }
    else {
        if ($runner eq '0') {
            &debug_msg("Websocket Starter: Init \n");
            &vzlist;
            &parse_vzlist;
            &launcher;
            &debug_msg("Websocket Starter: Done \n");
        }
    }

    usleep $sleep;
    $firstrun = '0';
    $runner++;

    # After a few watchdog events reset the counter:
    if ($runner gt '4') {
        $runner = '0';
    }
}

#
### Subs
#

sub launcher {
    foreach my $key (keys %$vnc_users) {
        $PortNumber = $vnc_users->{$key}->{'port'};
        if (defined $PortNumber) {
            $check = `/usr/bin/pgrep -f "/websockify/run(.*):$PortNumber"`;
            chomp ($check);
            @checkPIDs = (split(/\n/,$check));
        }

        if ((($vnc_users->{$key}->{'mode'} eq 'manual') || ($vnc_users->{$key}->{'mode'} eq 'auto')) && ($vnc_users->{$key}->{'State'} eq 'running')) {
            # VNC should be on.
            $numelements = @checkPIDs;
            if ($numelements gt '0') {
                &debug_msg("VNC websocket for VPS $key is listening on port $PortNumber \n");
            }
            else {
                &debug_msg("VNC websocket for VPS $key is NOT listening on port $PortNumber. Restarting Websocket client. \n");
                # Start and fork websocket launcher:
                unless ($pid = fork) {
                    unless (fork) {
                        exec "/usr/sausalito/novnc/utils/launch.sh --cert /etc/admserv/certs/combined --web /usr/sausalito/ui/web/.novnc/ --ssl-only --vnc ::1:$PortNumber --listen 1$PortNumber >/dev/null 2>&1";
                        &debug_msg("ERROR: VNC websocket for VPS $key for port $PortNumber failed to start!\n");
                    }
                }
                waitpid($pid, 0);
            }
        }

        # Reset/Undef Vars:
        undef $PortNumber;
        undef $check;
        undef $checkPIDs;
    }

}

sub parse_vzlist {
    # Find out which CTs and VMs have VNC enabled and on which port:
    foreach my $key (@$vzlist) {
        if ($key->{'uuid'} ne "") {
            $vzinfo_raw = `/usr/bin/prlctl list -ij $key->{'uuid'}`;
            chomp($vzinfo_raw);
            $vpsinfo = from_json($vzinfo_raw, { utf8  => 1 } );
            $uuid = $key->{'uuid'};
            foreach my $vkey (@$vpsinfo) {
                if ($vkey->{'Remote display'}->{'mode'} ne "off") {
                    if ((defined($vkey->{'Remote display'}->{'mode'})) && (defined($vkey->{'Remote display'}->{'port'}))) {
                        $vnc_users->{$uuid}->{'State'} = $vkey->{'State'};
                        $vnc_users->{$uuid}->{'mode'} = $vkey->{'Remote display'}->{'mode'};
                        $vnc_users->{$uuid}->{'port'} = $vkey->{'Remote display'}->{'port'};
                        undef $vkey->{'Remote display'};                                
                    }
                }
                else {
                    $vnc_users->{$uuid}->{'State'} = $vkey->{'State'};
                    $vnc_users->{$uuid}->{'mode'} = $vkey->{'Remote display'}->{'mode'};
                }
            }
        }
    }
}

sub vzlist {
    # Get list of CTs and VMs:
    $vzlist_raw = `/usr/bin/prlctl list -aj`;
    chomp($vzlist_raw);
    $vzlist = from_json($vzlist_raw, { utf8  => 1 } );
}

sub debug_msg {
    if ($DEBUG) {
        my $msg = shift;
        $user = $ENV{'USER'};
        setlogsock('unix');
        openlog($0,'','user');
        syslog('info', "$ARGV[0]: $msg");
        closelog;
    }
}

# 
# Copyright (c) 2018 Michael Stauber, SOLARSPEED.NET
# All Rights Reserved.
# 
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in 
#    the documentation and/or other materials provided with the 
#    distribution.
# 
# 3. Neither the name of the copyright holder nor the names of its 
#    contributors may be used to endorse or promote products derived 
#    from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE 
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
# 
# You acknowledge that this software is not designed or intended for 
# use in the design, construction, operation or maintenance of any 
# nuclear facility.
# 